#ifndef GRAPH_HPP
#define GRAPH_HPP

#include <map>
#include <string>
#include <vector>
#include <stdio.h>
#include <assert.h>
#include <iostream>
#include <stdlib.h>
#include "config.h"
// some trickery to get uint64_t --- you would think the natural
// way to do this would be to check for csdtint in the configure
// script and then just include it, however, cstdint only works
// with -std=c++0x at the moment --- this would crush all hopes
// for portable code

#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif


const char T_SEND = 1;
const char T_RECV = 2;
const char T_SENDVEC = 3;
const char T_RECVVEC = 4;
const char T_OP = 5;
const char T_OPVEC = 6;

const char MEM_UNDEFINED = 9;
const char MEM_USER = 10;
const char MEM_SCRATCH = 11;

class Node {
	private:
	uint64_t node_key; // to support all cases we need 32 bits for the rank and 32 bits for the offset
	void* tuple_loc;
	int tuple_len;
	int tuple_peer;
	int tuple_tag;
	int node_type;
	int depcnt; // for internal use of matcher, otherwise we have to modify NodesWeDependOn
	int tuple_rank; //owner
	char tuple_memtype;

	// We could also use ActiveQueues here...
	std::vector<Node*> NodesWhoDependOnUs;
	std::vector<Node*> NodesWeDependOn;

	// The matching send or recv
	Node* MatchingNode;

	public:

	Node(char nodetype, uint64_t nodekey, int rank, int len, char memtype, void* loc, int peer, int tag) {
		this->node_type = nodetype;
		this->node_key = nodekey;
		this->tuple_rank = rank;
		this->tuple_len = len;
		this->tuple_memtype = memtype;
		this->tuple_loc = loc;
		this->tuple_peer = peer;
		this->tuple_tag = tag;
	}

	void addNewChild(Node* n) {
	//	NodesWhoDependOnUs.insert(std::pair<int, Node*>(n->getKey(), n));
		NodesWhoDependOnUs.push_back(n);
	} 

	void addNewParent(Node* n) {
	//	NodesWeDependOn.insert(std::pair<int, Node*>(n->getKey(), n));
		NodesWeDependOn.push_back(n);
	}

	void printNode(void) {
		char memtype = 'U';
		if (this->getMemtype() == MEM_SCRATCH) memtype = 'S';
		std::cout << "Node (" 	<< this->getOwner()  << ":" 
								<< this->getType() << ":" 
								<< this->getPeer() << ":" 
								<< memtype << "[" << this->getAddr() << "]:" 
								<< this->getSize() << ":"
							    << this->getTag() << ")" << std::endl;  
	}

	std::string toString() {
		char buffer[1024];
		std::string type = "S";
		if (this->getType() == T_RECV) type = "R";
		snprintf(buffer, 1023, "%llu: %s %u", this->getKey(), type.c_str(), this->getOwner());
		return std::string(buffer);		
	}

	uint64_t getKey() { return this->node_key; }
	int getOwner(void) {return tuple_rank;}
	void* getAddr(void) {return tuple_loc;}
	char getMemtype(void) {return tuple_memtype;}
	int getSize(void) {return tuple_len;}
	int getPeer(void) {return tuple_peer;}
	int getTag(void) {return tuple_tag;}
	int getType(void) {return node_type;}

	void setDepCnt(int d) {this->depcnt = d;}
	int getDepCnt() {return this->depcnt;}
	void incrDepCnt() {this->depcnt += 1;}
	void decrDepCnt() {this->depcnt -= 1;}

	//std::map<int, Node*>* getNodesWhoDependOnUs(void) {
	std::vector<Node*>* getNodesWhoDependOnUs(void) {
		return &(this->NodesWhoDependOnUs);
	}

	//std::map<int, Node*>* getNodesWeDependOn(void) {
	std::vector<Node*>* getNodesWeDependOn(void) {
		return &(this->NodesWeDependOn);
	}

	void setMatchingNode(Node* match) { this->MatchingNode = match;	}
	Node* getMatchingNode() { return this->MatchingNode; }
};

class Graph {

	private:

	std::map<int, Node*> Nodes;
	std::map<int, Node*> ActiveNodes;

	public:

	~Graph() {
		std::map<int, Node*>::iterator it = Nodes.begin();
		while (it != Nodes.end()) {
			delete(it->second);
			it++;		
		}
		Nodes.clear();
		ActiveNodes.clear();		
	}

	void insertNewNode(Node* n) {
		n->setDepCnt(0);
		Nodes.insert(std::pair<int,Node*>(n->getKey(), n));
		ActiveNodes.insert(std::pair<int,Node*>(n->getKey(), n));
	}

	void addSend(uint64_t nodekey, int rank, int len, char memtype, void* loc, int peer, int tag) {
		Node* n = new Node(T_SEND, nodekey, rank, len, memtype, loc, peer, tag);
		this->insertNewNode(n);
	}
	
	void addRecv(uint64_t nodekey, int rank, int len, char memtype, void* loc, int peer, int tag) {
		Node* n = new Node(T_RECV, nodekey, rank, len, memtype, loc, peer, tag);
		this->insertNewNode(n);
	}

	void addDependency(uint64_t firstkey, uint64_t nextkey) {
		// get both nodes
		std::map<int, Node*>::iterator itf;
		itf = Nodes.find(firstkey);
		if (itf == Nodes.end()) {
			std::cout << "Woha, you tried to specify a dependency on an undefined node" << std::endl;
			exit(EXIT_SUCCESS);
		}
		std::map<int, Node*>::iterator itn;
		itn = Nodes.find(nextkey);
		if (itn == Nodes.end()) {
			std::cout << "Woha, you tried to specify a dependency on an undefined node" << std::endl;
			exit(EXIT_SUCCESS);
		}
		// remove nextkey node from ActiveNodes
		ActiveNodes.erase(nextkey);
		// add edges between them
		itf->second->addNewChild(itn->second);
		itn->second->addNewParent(itf->second);
		// increase nextkeys dependency counter
		itn->second->incrDepCnt();
	}

	std::map<int, Node*>* getActiveNodes(void) {
		return &ActiveNodes;
	}

};

class SetOfGraphs {
	std::map<int, Graph*> Graphs;

	public:

	~SetOfGraphs() {
		std::map<int, Graph*>::iterator it = Graphs.begin();
		while (it != Graphs.end()) {
			delete((it->second));
			it++;		
		}	
	}

	Graph* getGraphByRank(int rank) {
		// returns graph if it exist, otherwise create it
		std::map<int, Graph*>::iterator it;
		it = Graphs.find(rank);
		if (it == Graphs.end()) {
			Graph* g = new Graph;
			Graphs.insert(std::pair<int,Graph*>(rank, g));
			return g;
		}
		else {
			return it->second;
		}
	}

    typedef std::map<int, Graph*>::iterator iterator;
    typedef std::map<int, Graph*>::const_iterator const_iterator;

	iterator begin() { return Graphs.begin(); }
    const_iterator begin() const { return Graphs.begin(); }
    iterator end() { return Graphs.end(); }
    const_iterator end() const { return Graphs.end(); }

};

#endif

