#ifndef cDAG_TRAN_MPI_HPP
#define cDAG_TRAN_MPI_HPP
#include <mpi.h>
#include <stdarg.h>
#include "libcDAG.hpp"
#include "cDAG_Types.hpp"

#define cDAG_TRAN_MPI
#define NBC_DLEVEL 0

#if NBC_DLEVEL == 0 
#define NBC_DEBUG(a, ...) 
#else
static __inline__ void NBC_DEBUG(int level, const char *fmt, ...) { 
  va_list ap;
  int rank; 
 
  if(NBC_DLEVEL >= level) { 
    MPI_Comm_rank(MPI_COMM_WORLD, &rank); 
    
    printf("[LibNBC - %i] ", rank); 
    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end (ap);
  } 
}
#endif

#define T_SEND 1
#define T_RECV 2 
#define T_OP 3

/* the send argument struct */
typedef struct {
  void *buf;
  cDAG_MemType memtype;
  int count;
  int dest;
} NBC_Args_send;

/* the receive argument struct */
typedef struct {
  void *buf;
  cDAG_MemType memtype;
  int count;
  int source;
} NBC_Args_recv;

/* the operation argument struct */
typedef struct {
	int opnum;
	void (*func)(void*, void*);
	int argsize;
	void* args; // this is not! a pointer to the args, the args are stored here, they are of lenght argsize
} NBC_Args_op;

#define SCHED_ERR -2
#define SCHED_NONE -1
#define SCHED_DONE 1
#define SCHED_CONTINUE 2

static inline int type_get_size(char type) {
  switch(type) {
    case T_SEND:  return sizeof(NBC_Args_send);
                  break;
    case T_RECV:  return sizeof(NBC_Args_recv);
                  break;
    case T_OP:    return sizeof(NBC_Args_op);
                  break;
    default:
                  printf("[type_get_size] type %i not supported!\n", (int)type);
                  return 0;
                  break;
  }
  return -1;
}

struct bcast_args {
	void* buffer;
	int count;
	int root;
};

struct scatter_args {
	void* sendbuffer;
	void* recvbuffer;
	int count;
	int root;
};

struct alltoall_args {
	void* sendbuffer;
	void* recvbuffer;
	int count;
};

static inline int predefined_op_get_argsize(int opnum) {
	switch (opnum) {

		case -1:
			return 0;
			break;
		case -2:
			return sizeof(struct bcast_args);
			break;
		case -3:
			return sizeof(struct scatter_args);
			break;
		case -4:
			return sizeof(struct scatter_args);
			break;
		case -5:
			return sizeof(struct alltoall_args);
			break;
		default:
			return 0;
	}
}

static inline void execute_predefined_op(int opnum, void* args, void* scratch) {
	if (opnum == -1) {
		MPI_Barrier(G_cDAG_WorldComm);
	}
	else if (opnum == -2) {
		struct bcast_args* bc = (struct bcast_args*) args;
		MPI_Bcast(bc->buffer, bc->count, MPI_BYTE, bc->root, G_cDAG_WorldComm);
	}
	else if (opnum == -3) {
		struct scatter_args* sc = (struct scatter_args*) args;
		MPI_Scatter(sc->sendbuffer, sc->count, MPI_BYTE, sc->recvbuffer, sc->count, MPI_BYTE, sc->root, G_cDAG_WorldComm);
	}
	else if (opnum == -4) {
		struct scatter_args* ga = (struct scatter_args*) args;
		MPI_Gather(ga->sendbuffer, ga->count, MPI_BYTE, ga->recvbuffer, ga->count, MPI_BYTE, ga->root, G_cDAG_WorldComm);
	}
	else if (opnum == -5) {
		struct alltoall_args* aa = (struct alltoall_args*) args;
		MPI_Alltoall(aa->sendbuffer, aa->count, MPI_BYTE, aa->recvbuffer, aa->count, MPI_BYTE, G_cDAG_WorldComm);
	}
	else  {
		printf("Predefined op number %i is not implemented yet\n", opnum);
	}
}

/* executes element and returns SCHED_DONE if th operation is finished
 * and SCHED_CONTINUE and a request handle if the operations is running */
static inline int execute_element(char type, void *val, sched_req *req, cDAG_Handle handle) {
	switch(type) {
	case T_SEND:  {
		NBC_Args_send *args=(NBC_Args_send*) val;
		NBC_DEBUG(5,"SEND (val: %lu) *buf: %p, count: %i, dest: %i, tag: %i)\n", (unsigned long)val, args->buf, args->count, args->dest, handle->tag);
		void *buf1;
		if(args->memtype == cDAG_SCRATCHPAD)
			buf1=(char*)handle->tmpbuf+(long)args->buf;
		else
			buf1=args->buf;
		int res = MPI_Isend(buf1, args->count, MPI_BYTE, args->dest, handle->tag, G_cDAG_WorldComm, req);
		if(MPI_SUCCESS != res) { printf("Error in MPI_Isend(%lu, %i, %lu, %i, %i, %lu) (%i)\n", (unsigned long)buf1, args->count, 0, args->dest, handle->tag, (unsigned long)G_cDAG_WorldComm, res); goto error; }
		return SCHED_CONTINUE;
		break;
	}
	case T_RECV:  {
		NBC_Args_recv *args=(NBC_Args_recv*) val;
		NBC_DEBUG(5,"RECV (val: %lu) *buf: %p, count: %i, src: %i, tag: %i)\n", (unsigned long)val, args->buf, args->count, args->source, handle->tag);
		void *buf1;
		if(args->memtype == cDAG_SCRATCHPAD)
			buf1=(char*)handle->tmpbuf+(long)args->buf;
		else
			buf1=args->buf;
		int res = MPI_Irecv(buf1, args->count, MPI_BYTE, args->source, handle->tag, G_cDAG_WorldComm, req);
		if(MPI_SUCCESS != res) { printf("Error in MPI_Irecv(%lu, %i, %lu, %i, %i, %lu) (%i)\n", (unsigned long)buf1, args->count, 0, args->source, handle->tag, (unsigned long)G_cDAG_WorldComm, res); goto error; }
		return SCHED_CONTINUE;
		break;
	}
	case T_OP: {
		NBC_Args_op* args = (NBC_Args_op*) val;
		if (args->opnum < 0) execute_predefined_op(args->opnum, &args->args, handle->tmpbuf);
		else args->func(&args->args, handle->tmpbuf);
		return SCHED_CONTINUE;
		break;
	}
	default:
		printf("[print_type_content] type %i not supported!\n", (int)type);
		break;
	}

	error:
	return SCHED_ERR;
}

/* tests for outstanding elements and returns the finished index or
 * SCHED_NONE  */
static inline int scheduler_testany(sched_req *reqs, int count) {

  int index, flag;
  int ret = MPI_Testany(count, reqs, &index, &flag, MPI_STATUS_IGNORE);
  if(flag) {
    return index;
  } else {
    return SCHED_NONE;
  }
}

static inline int scheduler_test(sched_req req) {
  int flag;
  MPI_Test(&req, &flag, MPI_STATUS_IGNORE);
  if(flag) return SCHED_DONE;
  else return SCHED_NONE;
}

/* waits for outstanding elements and returns the finished index */
static inline int scheduler_waitany(sched_req *reqs, int count) {

  int index;
  int ret = MPI_Waitany(count, reqs, &index, MPI_STATUS_IGNORE);
  return index;
}


#endif
