#include <mpi.h>
#include "config.h"
#include "Graph.hpp"
#include "libcDAG.hpp"
#include "Matcher.hpp"
#include "DataFlow.hpp"
#include "HtorSets.hpp"
#include "cDAG_Types.hpp"
#include "cDAG_Scheduler.hpp"
#include "DetectCollectives.hpp"

/**
 * This function bla
 */

void unpack_binary_schedules(int num_schedules, schedule_t* schedules, int* sched_sizes, SetOfGraphs* LocalGraphs) {

	for (uint64_t rank=0; rank<num_schedules; rank++) { ///< needs to be 64 bit because we shift it and append local node id
		Graph* graph = LocalGraphs->getGraphByRank(rank);
		int offset = 0;
		// parsing is done in two passes, first we add all vertices
		// in the second run we take care of the edges
		offset += sizeof(int); // jump over scratchpad size
		offset += sizeof(int); // jump over the number of independent actions
		while (offset < sched_sizes[rank]) {
			int node_start_offset = offset;
			char type;
			SCHED_GET(&type, schedules[rank], offset, sizeof(char));
			offset += sizeof(int); // jump over dependency counter
			int num_outedges;
			SCHED_GET(&num_outedges, schedules[rank], offset, sizeof(int));
			offset += num_outedges * sizeof(int);
			NBC_Args_send *args_s;
			NBC_Args_recv *args_r;
			switch (type) {
				case T_SEND:
					args_s = (NBC_Args_send*) (schedules[rank] + offset);
					graph->addSend((rank << 32) | node_start_offset, rank, args_s->count, args_s->memtype, args_s->buf, args_s->dest, /*tag*/ 0);
					offset += sizeof(NBC_Args_send);
					break;
				case T_RECV:
					args_r = (NBC_Args_recv*) (schedules[rank] + offset);
					graph->addRecv((rank << 32) | node_start_offset, rank, args_r->count, args_r->memtype, args_r->buf, args_r->source, /*tag*/ 0);
					offset += sizeof(NBC_Args_recv);
					break;
				default:
					printf("*** Type %u not supported ***\n", type);
					exit(EXIT_FAILURE);
			}
		}

		// this is the second pass where we add all the edges
		offset = 0;
		offset += sizeof(int); // jump over scratchpad size
		offset += sizeof(int); // jump over the number of independent actions
		while (offset < sched_sizes[rank]) {
			int node_start_offset = offset;
			char type;
			SCHED_GET(&type, schedules[rank], offset, sizeof(char));
			offset += sizeof(int); // jump over dependency counter
			int num_outedges;
			SCHED_GET(&num_outedges, schedules[rank], offset, sizeof(int));
			for (int i=0; i < num_outedges; i++) {
				int outedge_to;
				SCHED_GET(&outedge_to, schedules[rank], offset, sizeof(int));
				graph->addDependency((rank << 32) | node_start_offset, (rank << 32) | outedge_to);
			}
			switch (type) {
				case T_SEND:
					offset += sizeof(NBC_Args_send);
					break;
				case T_RECV:
					offset += sizeof(NBC_Args_recv);
					break;
				default:
					printf("*** Type %u not supported ***\n", type);
					exit(EXIT_FAILURE);
			}
		}
	}

}

void print_tuples(std::vector<tuple*>* tuples) {
	for(std::vector<tuple*>::iterator ri = tuples->begin(); ri!=tuples->end(); ++ri) {
		tuple *t = *ri;
	}
}

void generate_goalop_for_collective(Collective* coll, cDAG_Graph* new_graphs) {

	print_tuples(coll->tuples);

	switch (coll->colltype) {
		case BCAST: {
			struct bcast_args bca;
			for (int i=0; i<coll->tuples->size(); i++) {
				tuple* t = coll->tuples->at(i);
				bca.buffer = (void*) t->start;
				bca.count = t->size;
				bca.root = t->sowner;
				cDAG_Op(new_graphs[t->owner], -2 /*bcast opcode, see Tran_MPI*/, &bca);
			}
			delete coll;
			break;
		}

		case SCATTER: {
			struct scatter_args sca;
			for (int i=0; i<coll->tuples->size(); i++) {
				tuple* t = coll->tuples->at(i);
				sca.sendbuffer = (void*) t->sstart;
				sca.recvbuffer = (void*) t->start;
				sca.count = t->size;
				sca.root = t->sowner;
				cDAG_Op(new_graphs[t->owner], -3 /*scatter opcode, see Tran_MPI*/, &sca);
			}
			delete coll;
			break;
		}

		case GATHER: {
			struct scatter_args gaa;
			for (int i=0; i<coll->tuples->size(); i++) {
				tuple* t = coll->tuples->at(i);
				gaa.sendbuffer = (void*) t->sstart;
				gaa.recvbuffer = (void*) t->start;
				gaa.count = t->size;
				gaa.root = t->owner;
				cDAG_Op(new_graphs[t->sowner], -4 /*gather opcode, see Tran_MPI*/, &gaa);
			}
			delete coll;
			break;
		}

		case ALLTOALL: {
			struct alltoall_args aaa;
			for (int i=0; i<coll->tuples->size(); i++) {
				tuple* t = coll->tuples->at(i);
				if (t->owner == t->sowner) {
					aaa.sendbuffer = (void*) (t->sstart - t->size * t->sowner);
					aaa.recvbuffer = (void*) (t->start - t->size * t->owner);
					aaa.count = t->size;
					cDAG_Op(new_graphs[t->sowner], -5 /*gather opcode, see Tran_MPI*/, &aaa);
				}
			}
			delete coll;
			break;
		}

		default:
			printf("Collective not implemented yet\n");
			break;
	}

}

void perform_collective_detection(cDAG_Schedule* sched) {
	
	schedule_t* schedules = (schedule_t*) malloc(G_cDAG_GroupSize * sizeof(schedule_t)); // pointers to local schedules
	int* sched_sizes = (int*) malloc(G_cDAG_GroupSize * sizeof(int)); // array of schedule sizes
	MPI_Gather(&sched->size, 1, MPI_INT, sched_sizes, 1, MPI_INT, 0, G_cDAG_WorldComm);
	int sum = 0; // for calculating displacements
	int* displs = (int*) malloc(G_cDAG_GroupSize * sizeof(int));
	for (int i=0; i<G_cDAG_GroupSize; i++) {
		displs[i]=sum;
		sum += sched_sizes[i];
	}
	char* schedules_real = (char*) malloc(sum);
	int mem_allocated = sum;
	for (int i=0; i<G_cDAG_GroupSize; i++) schedules[i] = (schedule_t) (((char*) schedules_real) + displs[i]);
	MPI_Gatherv(sched->active, sched->size, MPI_BYTE, schedules_real, sched_sizes, displs, MPI_BYTE, 0, G_cDAG_WorldComm);
	free(sched->active);

#ifdef HAVE_PGT
	PGT_Rec(3); //record gather id:3
#endif

	int newsched_size = 0;

	if (G_cDAG_MyRank == 0) {

		SetOfGraphs LocalGraphs;
		unpack_binary_schedules(G_cDAG_GroupSize, schedules, sched_sizes, &LocalGraphs); // parses the binary schedules and fills LocalGraphs
#ifdef HAVE_PGT
		PGT_Rec(4); //record unpack id:4 
#endif
		HtorSets* hs = Matcher(&LocalGraphs);
#ifdef HAVE_PGT
		PGT_Rec(5); //record matching id:5 
#endif
	//	hs->printGlobalGraphAsDot();
		std::vector<tuple*>* tuples = DataFlowAnalysis(hs);
		delete hs;
#ifdef HAVE_PGT
		PGT_Rec(6); //record data flow analysis id:6
#endif
	//	printf("old tuples\n");
	//	print_tuples(tuples);
		std::vector<Collective*> colls;
		tuples = DetectCollectives(G_cDAG_GroupSize, tuples, &colls);
#ifdef HAVE_PGT
		PGT_Rec(7); //record collective detection id:7
#endif
	//	printf("new tuples\n");
	//	print_tuples(tuples);
		
		// create new graphs based on the collectives that have been found
		cDAG_Graph* new_graphs = (cDAG_Graph*) malloc(sizeof(cDAG_Graph) * G_cDAG_GroupSize);
		for (int rank=0; rank < G_cDAG_GroupSize; rank++) new_graphs[rank] = _cDAG_Create_graph_internal();
		// add all detected collectives to the new graphs (as predefined operations which make MPI calls)
		printf("found %i collectives\n", colls.size());
		for (int c=0; c<colls.size(); c++) generate_goalop_for_collective(colls[c], new_graphs);
		// now add send-recv pairs for each remaining tuple (a remaining tuple is a tuple not replaced by a coll)
		for (std::vector<tuple*>::iterator ri = tuples->begin(); ri!=tuples->end(); ++ri) {
			tuple *t = *ri;
			cDAG_Send(new_graphs[t->sowner], (void*) t->sstart, t->ssize, t->owner); // TODO we need tags here!
			cDAG_Recv(new_graphs[t->owner], (void*) t->start, t->size, t->sowner); // TODO we need tags here!
			free(t);
		}
		delete tuples;

#ifdef HAVE_PGT
		PGT_Rec(8); //record building of new graphs id:8
#endif

		int size_acc = 0;
		
		// compile the newly created graphs into binary schedules
		// set do_coll_detection to 0 to avaoid recursion!
		for (int rank=0; rank < G_cDAG_GroupSize; rank++) {
			displs[rank] = size_acc;
			new_graphs[rank]->do_coll_detection = 0;
			cDAG_Schedule nsched = cDAG_Compile_internal(new_graphs[rank]);
			cDAG_FreeGraph(new_graphs[rank]);
			if (mem_allocated == 0) {
				mem_allocated = nsched.size;
				schedules_real = (char*) malloc(mem_allocated);
			}
			while (mem_allocated < (size_acc + nsched.size)) {
				mem_allocated =  mem_allocated *= 2;
				schedules_real = (char*) realloc(schedules_real, mem_allocated);
			}
			memcpy(schedules_real + size_acc, nsched.active, nsched.size);
			sched_sizes[rank] = nsched.size;
			size_acc += sched_sizes[rank];
			cDAG_FreeSchedule(nsched);
		}
		free(new_graphs);

#ifdef HAVE_PGT
		PGT_Rec(9); //record compiling of new graphs id:9
#endif
	}

	MPI_Scatter(sched_sizes, 1, MPI_INT, &newsched_size, 1, MPI_INT, 0, G_cDAG_WorldComm);
	char* newsched = (char*) malloc(newsched_size);
	MPI_Scatterv(schedules_real, sched_sizes, displs, MPI_BYTE, newsched, newsched_size, MPI_BYTE, 0, G_cDAG_WorldComm);
	sched->active = newsched;
	sched->size = newsched_size;

	free(sched_sizes);
	free(schedules);
	free(schedules_real);
	free(displs);

#ifdef HAVE_PGT
	PGT_Rec(10); //record scatter of new graphs id:10
#endif
}

