#include <mpi.h>
#include "tests.hpp"
#include "libcDAG.hpp"

void init_buffer(void* buffer, size_t buffer_size) {
	
	srand(time(NULL));
	
	for (size_t i=0; i<buffer_size; i++) {
		char rnd = rand() % 256;
		((char*) buffer)[i] = rnd;
	}

}

int* figure_out_random_peers() {

	const float msg_xtra = 1.0;

	int p, r;
	MPI_Comm_size(MPI_COMM_WORLD, &p);
	MPI_Comm_rank(MPI_COMM_WORLD, &r);

	/*
	We want msg_xtra*p extra messages in the network, between random nodes
	of course each send needs a matching recv, so figure out who will
	communicate with who and tell the peers about that
	*/

	int* msgs = (int*) calloc(p*p,sizeof(int));

	if (r == 0) {
		srand(time(NULL));
		for (int i=0; i<p; i++) {
			int sender = 0;
			int receiver = 0;
			while (sender == receiver) {
				sender = rand() % p;
				receiver = rand() % p;
			}
			msgs[receiver*p+sender]++;
		}	
	}

	// That's a waste of bandwidth to bcast the whole array, but it's not time critical anyway
	MPI_Bcast(msgs, p*p, MPI_INT, 0, MPI_COMM_WORLD);

	return msgs;
		
}

static inline int inorder_binomial_tree_father(int rank, int size, int root) {

	int mask = 1;

	if ((rank < 0) || (rank > size) || (size < 1)) return -1; // sanity check

	int vrank = (rank - root + size) % size;

	if (root == rank) {
		return root;
	}

	while (mask < size) {
		int remote = vrank ^ mask;
		if (remote < vrank) {
			return ((remote + root) % size);
		}
		mask <<= 1;
	}

	return -1;
}

static inline int inorder_binomial_tree_child(int rank, int size, int num_child, int root) {

	int childs = 0;
	int mask = 1;

	int vrank = (rank - root + size) % size;

	while (mask < size) {
		int remote = vrank ^ mask;
		if (remote < vrank) break;
		else if (remote < size) {
			if (childs==num_child) return ((remote + root) % size);
			childs++;
		}
		mask <<= 1;
	}

	return -1;
}

int main(int argc, char** argv) {

	MPI_Init(&argc, &argv);
	TEST_INIT("binomial broadcast");

	cDAG_Init();

	const int datasize = 8;
	int r = cDAG_MyRank();
	int p = cDAG_GroupSize();
	char* buffer_cdag = (char*) malloc(datasize);
	char* buffer_mpi = (char*) malloc(datasize);
	char* random_buffer = (char*) malloc(2*datasize);
	init_buffer(buffer_cdag, datasize);
	init_buffer(buffer_mpi, datasize);
	init_buffer(random_buffer, 2*datasize);

	int* msgs = figure_out_random_peers();

for (int i=0; i<1000; i++) {
	cDAG_Graph g = cDAG_Create_graph();

	if (r == 0) {
		/* we are root, send data to all our children */
		int childnum = 0;
		int child;
		while ((child = inorder_binomial_tree_child(r, p, childnum, 0)) != -1) {
			cDAG_Send(g, buffer_cdag, datasize, child);
			printf("[rank %i] send to %i\n", r, child);
			childnum++;
		}
	}
	else {
		/* we are a child, recv data from father and send it to all children after that*/
		int father = inorder_binomial_tree_father(r, p, 0);
		cDAG_Vertex recv = cDAG_Recv(g, buffer_cdag, datasize, father);
		int childnum = 0;
		int child;
		while ((child = inorder_binomial_tree_child(r, p, childnum, 0)) != -1) {
			cDAG_Vertex send = cDAG_Send(g, buffer_cdag, datasize, child);
			cDAG_Requires(g, send, recv);
			childnum++;
		}
	
	}

	/* the random msgs are added now (sends)*/
	for (int peer=0; peer<p; peer++) {
		int n_sends = msgs[peer*p+r];
		for (int m=0; m<n_sends; m++) { // note that n_sends should be either 0 (most cases) or 1, however the possibility of it beeing > 1 is not null
			cDAG_Send(g, random_buffer, datasize*2, peer);
		}
		int n_recvs = msgs[r*p+peer];
		for (int m=0; m<n_recvs; m++) {
			cDAG_Recv(g, random_buffer, datasize*2, peer);
		}
	}

	g->do_coll_detection = 0;
	cDAG_Schedule sched = cDAG_Compile(g);
	cDAG_FreeGraph(g);
	cDAG_Handle hndl = cDAG_Run(sched);
	cDAG_Wait(hndl);
	cDAG_FreeSchedule(sched);

}

	MPI_Bcast(buffer_mpi, datasize, MPI_BYTE, 0, MPI_COMM_WORLD);

	if (memcmp(buffer_mpi, buffer_cdag, datasize) != 0) TEST_FAILED();
	else TEST_SUCCESSFULL();

	cDAG_Finalize();
	MPI_Finalize();

}

